/* work.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <assert.h>
#include <math.h>
#include <stdio.h>
#include "globals.h"
#include "grid.h"
#include "memory.h"
#include "queue.h"
#include "sync.h"
#include "traj.h"


#define DEG2RAD(d)  ((d)*3.14159/180.0)
#define MIN(A,B)  ( (A) < (B) ? (A) : (B) )
#define MAX(A,B)  ( (A) > (B) ? (A) : (B) )



/** function prototypes **/
void calc_surface();
void calc_hslice();
void calc_vslice();
void calc_chslice();
void calc_cvslice();
void calc_hwindslice();
void calc_vwindslice();
void calc_traj();
void compress_vertices();
void compress_normals();
void compress_float_to_char();
void int8_to_int4();




void work( tasknum )
int tasknum;
{
   int type;
   int i1, i2, i3;
   float f1, f2, f3, f4, f5;
   int time, parm;

   if (Debug) printf("work(%d)  %x\n", tasknum, &time );

   while (1) {
      get_qentry( &type, &i1, &i2, &i3, &f1, &f2, &f3, &f4, &f5 );
      if (Debug) printf("got entry: %d\n", type );
      time = i1;
      parm = i2;
      if (Debug)  printf("\ntask type=%d\n", type );
      switch (type) {
         case TASK_ISOSURFACE:
            /* compute an isosurface. */
            calc_surface( time, parm, IsoLevel[parm], tasknum );
	    break;
	 case TASK_HSLICE:
	    /* calculate a horizontal contour line slice. */
	    calc_hslice( time, parm,
			HSliceInterval[parm], HSliceLevel[parm], tasknum );
	    break;
	 case TASK_VSLICE:
	    /* calculate a horizontal contour line slice. */
	    calc_vslice( time, parm, VSliceInterval[parm],
			 VSliceR1[parm], VSliceC1[parm],
			 VSliceR2[parm], VSliceC2[parm], tasknum );
	    break;
	 case TASK_CHSLICE:
	    /* calculate a horizontal colored slice */
	    calc_chslice( time, parm, CHSliceLevel[parm], tasknum );
	    break;
	 case TASK_CVSLICE:
	    /* calculate a vertical colored slice */
	    calc_cvslice( time, parm, CVSliceR1[parm], CVSliceC1[parm],
			 CVSliceR2[parm], CVSliceC2[parm], tasknum );
	    break;
	 case TASK_HWIND:
	    /* calculate a horizontal wind slice */
	    calc_hwindslice( time, parm, HWindLevel[parm],
			    HWindScale[parm], HWindDensity[parm], tasknum );
	    break;
	 case TASK_VWIND:
	    /* calculate a vertical wind slice */
	    calc_vwindslice( time, parm, VWindR1[parm], VWindC1[parm],
			    VWindR2[parm], VWindC2[parm],
			    VWindScale[parm], VWindDensity[parm], tasknum );
	    break;
	 case TASK_TRAJ:
	    /* calculate a wind trajectory */
	    calc_traj( f1, f2, f3, time, i2, i3, f4, f5 );
	    break;
	 case TASK_QUIT:
	    if (Debug)
	       printf("TASK_QUIT\n");
	    return;
	    break;
         case TASK_NULL:
	    break;
	 default:
            printf("ERROR:  Undefined task code!!\n");
      } /*switch*/
   }

}



#define MAX_VERTS 40000


/*** calc_surface *****************************************************
   Calculate an iso-level surface and store it.
   Input:  time - the time step.
           parm - the parameter.
           iso_level - level to construct contour at.
	   tasknum - which cpu to execute on.
   Output:  resulting poly-triangle strip is saved in SurfTable.
**********************************************************************/
void calc_surface( time, parm, iso_level, tasknum )
int time, parm;
float iso_level;
int tasknum;
{
   /* marching cubes parameters */
   int debug = 0;
   float arx=1.0, ary=1.0, arz=1.0;
   int max = MAX_VERTS;
   float vx[MAX_VERTS], vy[MAX_VERTS], vz[MAX_VERTS];
   float nx[MAX_VERTS], ny[MAX_VERTS], nz[MAX_VERTS];
   int max2 = 2*MAX_VERTS;
   int vpts[2*MAX_VERTS];
   int ivert, ipts, ipoly, itri;
   /* other vars */
   int i, j, realparm;
   float *grid;
   int verts, vbytes, nbytes, bytes;
   float xscale, yscale, zscale;
   int_2 *cvx, *cvy, *cvz;
   int_1 *cnx, *cny, *cnz;
   uint_2 *pts;

   if (Debug)
      printf("\ncalc_surface%d(%d,%d,%f)...\n",tasknum,time,parm,iso_level );

   realparm = CloneTable[parm];

   grid = get_grid( time, realparm );  /* read the grid */
   if (!grid) return;

   main_march( grid, Nc, Nr, Nl, iso_level,
           arx, ary, arz, max,
           vx,vy,vz, nx,ny,nz, max2, vpts,
           &ivert, &ipts, &ipoly, &itri );

   release_grid( time, realparm, grid );

   if (Debug)
      printf("ivert=%d  ipts=%d  ipoly=%d  itri=%d\n", ivert,ipts,ipoly,itri );
   /* ivert is the number of vertices calculated */
   /* ipoly is the number of polygons calculated */
   /* ipts is the size of the vpts array which is used to make the
           polytriangle strip from the vertices. */
   /* itri is the number of triangles calculated.  (usually 0 for some
           unknown reason) */

   recent( ISOSURF, parm );

   if (ipts>65000)
      ipts = 65000;

   /*************************** Compress data ***************************/

   /* allocate memory to store compressed vertices */
   if (ivert && ipts) {
      vbytes = ivert*sizeof(int_2);
      cvx = (int_2 *) allocate( vbytes );
      cvy = (int_2 *) allocate( vbytes );
      cvz = (int_2 *) allocate( vbytes );
   
      /* convert vertices from grid coords to integers in [-10000,10000] */
      compress_iso_verts2( vx, vy, vz, cvx, cvy, cvz, ivert );

      /* allocate memory to store compressed normals */
      nbytes = ivert*sizeof(int_1);
      cnx = (int_1 *) allocate( nbytes );
      cny = (int_1 *) allocate( nbytes );
      cnz = (int_1 *) allocate( nbytes );

      /* convert normals from floats in [-1,1] to 1-byte ints in [-125,125] */
      compress_iso_norms2( nx, ny, nz, cnx, cny, cnz, ivert );

      /* allocate memory to store indices array */
      bytes = ipts * sizeof(uint_2);
      pts = (uint_2 *) allocate( bytes );

      /* convert 4-byte vpts[] elements to 2-byte pts[] elements */
      for (i=0;i<ipts;i++)
	 pts[i] = (uint_2) vpts[i];

   }
   else {
      cvx = cvy = cvz = NULL;
      cnx = cny = cnz = NULL;
      pts = NULL;
      ivert = ipts = 0;
   }

   /************************ Store the new surface ************************/

   wait_lock( &SurfTable[parm][time].lock );

   /* deallocate existing surface, if any */
   free_isosurface( time, parm );

   /* add surface to table */
   SurfTable[parm][time].isolevel = iso_level;
   SurfTable[parm][time].numverts = ivert;
   SurfTable[parm][time].vx = cvx;
   SurfTable[parm][time].vy = cvy;
   SurfTable[parm][time].vz = cvz;
   SurfTable[parm][time].nx = cnx;
   SurfTable[parm][time].ny = cny;
   SurfTable[parm][time].nz = cnz;
   SurfTable[parm][time].numpts = ipts;
   SurfTable[parm][time].pts = pts;
   SurfTable[parm][time].valid = 1;
 
   clear_lock( &SurfTable[parm][time].lock );

   Redraw=1;
}




static int compress_iso_verts( vx, vy, vz, vpts, cvx, cvy, cvz, verts )
float vx[], vy[], vz[];
int vpts[];
int_2 cvx[], cvy[], cvz[];
int verts;
{
   float xscale, yscale, zscale;
   int i, j;

   xscale =  2.0 / (float) (Nc-1) * 10000.0;
   yscale = -2.0 / (float) (Nr-1) * 10000.0;
   zscale = -2.0 / (float) (Nl-1) * 10000.0;
   
   for (i=0;i<verts;i++) {   /* fully vectorized */
      j = vpts[i];
      cvx[i] = (int_2) (vx[j] * xscale - 10000.0);
      cvy[i] = (int_2) (vy[j] * yscale + 10000.0);
      cvz[i] = (int_2) (vz[j] * zscale + 10000.0);
   }
}


static int compress_iso_verts2( vx, vy, vz, cvx, cvy, cvz, verts )
float vx[], vy[], vz[];
int_2 cvx[], cvy[], cvz[];
int verts;
{
   float xscale, yscale, zscale;
   int i;

   xscale =  2.0 / (float) (Nc-1) * 10000.0;
   yscale = -2.0 / (float) (Nr-1) * 10000.0;
   zscale = -2.0 / (float) (Nl-1) * 10000.0;
   
   for (i=0;i<verts;i++) {   /* fully vectorized */
      cvx[i] = (int_2) (vx[i] * xscale - 10000.0);
      cvy[i] = (int_2) (vy[i] * yscale + 10000.0);
      cvz[i] = (int_2) (vz[i] * zscale + 10000.0);
   }
}


static int compress_iso_norms( nx, ny, nz, vpts, cnx, cny, cnz, verts )
float nx[], ny[], nz[];
int vpts[];
int_1 cnx[], cny[], cnz[];
int verts;
{
   int i, j;

   for (i=0;i<verts;i++) {   /* fully vectorized */
      j = vpts[i];
      cnx[i] = (int_1) (-nx[j] * 125.0);
      cny[i] = (int_1) ( ny[j] * 125.0);
      cnz[i] = (int_1) ( nz[j] * 125.0);
   }
}


static int compress_iso_norms2( nx, ny, nz, cnx, cny, cnz, verts )
float nx[], ny[], nz[];
int_1 cnx[], cny[], cnz[];
int verts;
{
   int i;

   for (i=0;i<verts;i++) {   /* fully vectorized */
      cnx[i] = (int_1) (-nx[i] * 125.0);
      cny[i] = (int_1) ( ny[i] * 125.0);
      cnz[i] = (int_1) ( nz[i] * 125.0);
   }
}


/*** extract_hslice ***************************************************
   Extract a horizontal slice from a 3-D grid.
   Input:  grid - pointer to Nl x Nr * Nc float array.
           level - position in [0,Nl-1] to extract slice from.
	   colmajor - 1 = column major, 0 = row major.
   Returned:  pointer to Nr x Nc array (slice) of floats
**********************************************************************/
static float* extract_hslice( grid, level, colmajor )
float *grid, level;
int colmajor;
{
   float *slice;
   int upper, lower, above, below;
   float a, b, g1, g2;
   int i, j;

   slice = (float *) allocate( Nr * Nc * sizeof(float) );

   /* extract the 2-D array from 3-D grid */
   lower = (int) level;
   upper = lower + 1;
   if (upper>Nl-1)
      upper = Nl - 1;
   a = level - (float) lower;
   b = 1.0 - a;
   below = lower * Nr * Nc;
   above = upper * Nr * Nc;

   /* interpolate between layers */
   if (colmajor) {
      for (j=0; j<Nc; j++) {
	 for (i=0; i<Nr; i++) {
	    g1 = grid[above+j*Nr+i];
	    g2 = grid[below+j*Nr+i];
	    if (g1>1.0e30 || g2>1.0e30) {
	       slice[j*Nr+Nr-i-1] = 1.0e35;
	    }
	    else {
	       slice[j*Nr+Nr-i-1] = a * g1 + b * g2; 
	    }
	 }
      }
   }
   else {
      /* change from column-major to row-major order */
      for (i=0; i<Nr; i++) {
	 for (j=0; j<Nc; j++) {
	    g1 = grid[above+j*Nr+i];
	    g2 = grid[below+j*Nr+i];
	    if (g1>1.0e30 || g2>1.0e30) {
	       slice[i*Nc+j] = 1.0e35;
	    }
	    else {
	       slice[i*Nc+j] = a * g1 + b * g2;
	    }
	 }
      }
   }

   return slice;
}




/*** extract_vslice ***************************************************
   Extract a vertical slice from a 3-D grid.
   Input:  grid - pointer to Nl x Nr * Nc float array.
           r1,c1,r2,c2 - position in [0,Nr-1],[0,Nc-1] to extract slice.
	   cols, rows - size of slice returned.
	   colmajor - 1 = column major, 0 = row major.
   Returned:  pointer to Nr x Nc array (slice) of floats
**********************************************************************/
static float* extract_vslice( grid, r1,c1,r2,c2, rows, cols, colmajor )
float *grid;
float r1,c1, r2,c2;
int rows, cols;
int colmajor;
{
   float gx,gy, stepx, stepy;
   int ic, ir, i, j, iii;
   float g1,g2,g3,g4, ei,ej;
   float *slice;

   /* allocate slice array */
   slice = (float *) allocate( rows * cols * sizeof(float) );
   if (!slice)
      die("out of memory in extract_vslice");

   /* initialize gx and gy- the real grid indices */
   gx = c1;
   gy = r1;

   /* calculate step sizes for iterating from c1 to c2 and r1 to r2 */
   stepx = (c2-c1) / (float) (cols-1);
   stepy = (r2-r1) / (float) (cols-1);

   if (colmajor) {

      for (ic=0; ic<cols; ic++) {
	 /* calculate error terms */
	 ei = gx - (float) ( (int) gx );   /* in [0,1) */
	 ej = gy - (float) ( (int) gy );   /* in [0,1) */
	 /* convert real gx,gy to integer i,j for array indexing */
	 i = (int) gx;
	 j = (int) gy;
	 if (i > Nc-1) {
	    i = Nc-1;
	 }
	 if (j > Nr-1) {
	    j = Nr-1;
	 }
	 for (ir=0; ir<rows; ir++) {
	    iii = ir * Nr * Nc;
	    /* fetch grid data */
	    g1 = grid[iii + i*Nr + j];          /*  g1 -- g2    +-- j */
	    g3 = grid[iii + i*Nr + j+1];        /*  |      |    |     */
	    if (ei!=0.0) {                      /*  g3 -- g4    i     */
	       g2 = grid[iii + (i+1)*Nr + j];
	       g4 = grid[iii + (i+1)*Nr + j+1];
	    }
	    else {
	       g2 = g4 = 0.0;
	    }
	    if (g1>1.0e30 || g2>1.0e30 || g3>1.0e30 || g4>1.0e30) {
	       slice[ ic*rows + ir ] = 1.0e35;
	    }
	    else {
	       slice[ ic*rows + ir ] = g1 * (1.0-ei) * (1.0-ej)
				     + g2 *      ei  * (1.0-ej)
				     + g3 * (1.0-ei) *      ej
				     + g4 *      ei  *      ej;
	    }
	 }
	 /* increment grid indices */
	 gx += stepx;
	 gy += stepy;
      }
   }
   else {
      for (ic=0; ic<cols; ic++) {
	 /* calculate error terms */
	 ei = gx - (float) ( (int) gx );   /* in [0,1) */
	 ej = gy - (float) ( (int) gy );   /* in [0,1) */
	 /* convert real gx,gy to integer i,j for array indexing */
	 i = (int) gx;
	 j = (int) gy;
	 if (i > Nc-1) {
	    i = Nc-1;
	 }
	 if (j > Nr-1) {
	    j = Nr-1;
	 }
	 for (ir=0; ir<rows; ir++) {
	    iii = ir * Nr * Nc;
	    /* fetch grid data */
	    g1 = grid[iii + i*Nr + j];            /* g1 -- g2    +-- j */
	    g3 = grid[iii + i*Nr + j+1];          /* |      |    |     */
	    if (ei!=0.0) {                        /* g3 -- g4    i     */
	       g2 = grid[iii + (i+1)*Nr + j];
	       g4 = grid[iii + (i+1)*Nr + j+1];
	    }
	    else {
	       g2 = g4 = 0.0;
	    }
	    if (g1>1.0e30 || g2>1.0e30 || g3>1.0e30 || g4>1.0e30) {
	       slice[ ir*cols + ic ] = 1.0e35;
	    }
	    else {
	       slice[ ir*cols + ic ] = g1 * (1.0-ei) * (1.0-ej)
		                     + g2 *      ei  * (1.0-ej)
	                    	     + g3 * (1.0-ei) *      ej
		                     + g4 *      ei  *      ej;
	    }
	 }
	 /* increment grid indices */
	 gx += stepx;
	 gy += stepy;
      }
   }

   return slice;
}





/*** calc_hslice ******************************************************
   Calculate an horizontal contour line slice and store it.
   Input:  time - the time step.
           parm - the parameter.
	   interval - interval between lines.
	   level - position of slice in [0..Nl-1].
	   tasknum - which cpu to execute on.
   Output:  resulting poly-triangle strip is saved in HSliceTable.
**********************************************************************/
void calc_hslice( time, parm, interval, level, tasknum )
int time, parm;
float interval, level;
int tasknum;
{
   float *grid;
   float *slice, *scratch;
   int upper, lower, above, below;
   float a, b, g1, g2;
   int i, j, ii, jj, ij;
   int num1, num2, num3, bytes;
   float vx1[MAX_VERTS],vy1[MAX_VERTS];  /* contour line vertices */
   float vx2[MAX_VERTS],vy2[MAX_VERTS];  /* hidden line vertices */
   float vx3[MAX_VERTS],vy3[MAX_VERTS];  /* label line vertices */
   int max = MAX_VERTS;
   char buffer[2*MAX_VERTS];
   float interval2;
   int realparm;
   int_2 *cvx1, *cvy1, *cvx2, *cvy2, *cvx3, *cvy3;

   if (Debug)
     printf("calc_hslice%d( %d, %d, %f, %f )\n",
	    tasknum, time,parm,interval,level);

   realparm = CloneTable[parm];

   grid = get_grid( time, realparm );  /* read the grid */
   if (!grid)
      die("couldn't get grid");

   slice = extract_hslice( grid, level, 1 );

   /* allocate scratch array */
   scratch = (float *) allocate( Nr * Nc * sizeof(int) );
   if (!scratch)
      die("Out of memory in calc_hslice");

   /* call contouring routine */
   interval2 = (float) interval;   /* solve float/double mixups */
#ifdef SINGLE_TASK
#  ifdef ibm
   zcontwxyz( slice, &Nr, &Nc, &interval2, scratch,
          vx1, vy1, &max, &num1,
	  vx2, vy2, &max, &num2,
	  vx3, vy3, &max, &num3
        );
#else
   zcontwxyz_( slice, &Nr, &Nc, &interval2, scratch,
          vx1, vy1, &max, &num1,
	  vx2, vy2, &max, &num2,
	  vx3, vy3, &max, &num3
        );
#  endif
#else
   if (tasknum==1)
      zcont1_( slice, &Nr, &Nc, &interval2, scratch,
	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
   else if (tasknum==2)
      zcont2_( slice, &Nr, &Nc, &interval2, scratch,
 	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
   else if (tasknum==3)
      zcont3_( slice, &Nr, &Nc, &interval2, scratch,
 	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
   else if (tasknum==4)
      zcont4_( slice, &Nr, &Nc, &interval2, scratch,
 	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
#endif

   release_grid( time, realparm, grid );

   deallocate( scratch, Nr * Nc * sizeof(int) );
   deallocate( slice, -1 );

   /* vertices are in [-1,1]^2 */


   if (Debug)
      printf("[%d] sending hslice vertices %d %d %d\n",
	     tasknum, num1,num2,num3);

   recent( HSLICE, parm );

   /*************************** Compress data ***************************/

   if (num1) {
      bytes = num1*sizeof(int_2);
      cvx1 = (int_2 *) allocate( bytes );
      cvy1 = (int_2 *) allocate( bytes );
      compress_vertices( vy1, cvx1, num1 );
      compress_vertices( vx1, cvy1, num1 );
   }
   else {
      cvx1 = cvy1 = NULL;
   }

   if (num2) {
      bytes = num2*sizeof(int_2);
      cvx2 = (int_2 *) allocate( bytes );
      cvy2 = (int_2 *) allocate( bytes );
      compress_vertices( vy2, cvx2, num2 );
      compress_vertices( vx2, cvy2, num2 );
   }
   else {
      cvx2 = cvy2 = NULL;
   }

   if (num3) {
      bytes = num3*sizeof(int_2);
      cvx3 = (int_2 *) allocate( bytes );
      cvy3 = (int_2 *) allocate( bytes );
      compress_vertices( vy3, cvx3, num3 );
      compress_vertices( vx3, cvy3, num3 );
   }
   else {
      cvx3 = cvy3 = NULL;
   }

   /************************ Store the new slice ************************/

   wait_lock( &HSliceTable[parm][time].lock );

   /* deallocate existing slice, if any */
   free_hslice( time, parm );

   /* store new slice */
   HSliceTable[parm][time].interval = interval;
   HSliceTable[parm][time].level = level;
   HSliceTable[parm][time].num1 = num1;
   HSliceTable[parm][time].vx1 = cvx1;
   HSliceTable[parm][time].vy1 = cvy1;
   HSliceTable[parm][time].num2 = num2;
   HSliceTable[parm][time].vx2 = cvx2;
   HSliceTable[parm][time].vy2 = cvy2;
   HSliceTable[parm][time].num3 = num3;
   HSliceTable[parm][time].vx3 = cvx3;
   HSliceTable[parm][time].vy3 = cvy3;
   HSliceTable[parm][time].valid = 1;

   clear_lock( &HSliceTable[parm][time].lock );

   Redraw=1;
}




/*** calc_vslice ******************************************************
   Calculate an vertical contour line slice and store it.
   Input:  time - the time step.
           parm - the parameter.
	   interval - interval between lines.
	   r1, c1 - position of 1st corner in [0..Nr-1],[0..Nc-1]
	   r2, c2 - position of 1st corner in [0..Nr-1],[0..Nc-1]
	   tasknum - which cpu to execute on.
   Output:  resulting poly-triangle strip is saved in VSliceTable.
**********************************************************************/
void calc_vslice( time, parm, interval, r1,c1, r2,c2, tasknum )
int time, parm;
float interval, r1,c1, r2,c2;
int tasknum;
{
   float *grid;
   float *slice, *scratch;
   int cols, rows;
   int i, j, ii, jj, ij;
   int num1, num2, num3, bytes;
   float vx1[MAX_VERTS],vy1[MAX_VERTS],vz1[MAX_VERTS];  /* contour lines */
   float vx2[MAX_VERTS],vy2[MAX_VERTS],vz2[MAX_VERTS];  /* hidden lines */
   float vx3[MAX_VERTS],vy3[MAX_VERTS],vz3[MAX_VERTS];  /* label lines */
   int max = MAX_VERTS;
   char buffer[2*MAX_VERTS];
   float x1,y1, x2,y2, dx, dy, r;
   float interval2;
   int realparm;
   int_2 *cvx1, *cvy1, *cvz1, *cvx2, *cvy2, *cvz2, *cvx3, *cvy3, *cvz3;

   realparm = CloneTable[parm];

   grid = get_grid( time, realparm );  /* read the grid */
   if (!grid)
      die("couldn't get grid");

   /* allocate 2-D array to contour, scratch array */
   rows = Nl;
   cols = (Nr>Nc) ? Nr : Nc;  /* kludge */

   slice = extract_vslice( grid, r1,c1,r2,c2, rows, cols, 1 );

   scratch = (float *) allocate( rows * cols * sizeof(int) );
   if (!scratch)
      die("Out of memory in calc_vslice");

   /* call contouring routine */
   interval2 = (float) interval;
#ifdef SINGLE_TASK
#  ifdef ibm
   zcontwxyz( slice, &rows, &cols, &interval2, scratch,
	  vx1, vy1, &max, &num1,
	  vx2, vy2, &max, &num2,
	  vx3, vy3, &max, &num3
        );
#  else
   zcontwxyz_( slice, &rows, &cols, &interval2, scratch,
	  vx1, vy1, &max, &num1,
	  vx2, vy2, &max, &num2,
	  vx3, vy3, &max, &num3
        );
#  endif
#else
  if (tasknum==1)
      zcont1_( slice, &rows, &cols, &interval2, scratch,
	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
   else if (tasknum==2)
      zcont2_( slice, &rows, &cols, &interval2, scratch,
 	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
   else if (tasknum==3)
      zcont3_( slice, &rows, &cols, &interval2, scratch,
 	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
   else if (tasknum==4)
      zcont4_( slice, &rows, &cols, &interval2, scratch,
 	      vx1, vy1, &max, &num1,
	      vx2, vy2, &max, &num2,
	      vx3, vy3, &max, &num3
	     );
#endif

   release_grid( time, realparm, grid );
   deallocate( scratch, rows * cols * sizeof(int) );
   deallocate( slice, -1 );

   /* convert 2-D line segments in [-1,1]^2 to 3-D coords in [-1,1]^3 */

   x1 = c1 / (Nc-1) * 2.0 - 1.0;
   y1 = r1 / (Nr-1) * 2.0 - 1.0;
   x2 = c2 / (Nc-1) * 2.0 - 1.0;
   y2 = r2 / (Nr-1) * 2.0 - 1.0;
   dx = (x2 - x1);
   dy = (y2 - y1);
   for (i=0;i<num1;i++) {
      /* stellar:  fully vectorized */
      r = (vy1[i] + 1.0) / 2.0;   /* r in [0,1] */
      vz1[i] = -vx1[i];
      vx1[i] = x1 + r * dx;
      vy1[i] = -(y1 + r * dy);
   }
   for (i=0;i<num2;i++) {
      /* stellar:  fully vectorized */
      r = (vy2[i] + 1.0) / 2.0;   /* r in [0,1] */
      vz2[i] = -vx2[i];
      vx2[i] = x1 + r * dx;
      vy2[i] = -(y1 + r * dy);
   }
   for (i=0;i<num3;i++) {
      /* stellar:  fully vectorized */
      r = (vy3[i] + 1.0) / 2.0;   /* r in [0,1] */
      vz3[i] = -vx3[i];
      vx3[i] = x1 + r * dx;
      vy3[i] = -(y1 + r * dy);
   }

   recent( VSLICE, parm );

   /*************************** Compress data ***************************/

   if (num1) {
      bytes = num1*sizeof(int_2);
      cvx1 = (int_2 *) allocate( bytes );
      cvy1 = (int_2 *) allocate( bytes );
      cvz1 = (int_2 *) allocate( bytes );
      compress_vertices( vx1, cvx1, num1 );
      compress_vertices( vy1, cvy1, num1 );
      compress_vertices( vz1, cvz1, num1 );
   }
   else {
      cvx1 = cvy1 = cvz1 = NULL;
   }

   if (num2) {
      bytes = num2*sizeof(int_2);
      cvx2 = (int_2 *) allocate( bytes );
      cvy2 = (int_2 *) allocate( bytes );
      cvz2 = (int_2 *) allocate( bytes );
      compress_vertices( vx2, cvx2, num2 );
      compress_vertices( vy2, cvy2, num2 );
      compress_vertices( vz2, cvz2, num2 );
   }
   else {
      cvx2 = cvy2 = cvz2 = NULL;
   }

   if (num3) {
      bytes = num3*sizeof(int_2);
      cvx3 = (int_2 *) allocate( bytes );
      cvy3 = (int_2 *) allocate( bytes );
      cvz3 = (int_2 *) allocate( bytes );
      compress_vertices( vx3, cvx3, num3 );
      compress_vertices( vy3, cvy3, num3 );
      compress_vertices( vz3, cvz3, num3 );
   }
   else {
      cvx3 = cvy3 = cvz3 = NULL;
   }

   /************************ Store the new slice ************************/

   wait_lock( &VSliceTable[parm][time].lock );

   /* deallocate existing slice, if any */
   free_vslice( time, parm );

   /* store new slice */
   VSliceTable[parm][time].interval = interval;
   VSliceTable[parm][time].r1 = r1;
   VSliceTable[parm][time].c1 = c1;
   VSliceTable[parm][time].r2 = r2;
   VSliceTable[parm][time].c2 = c2;
   VSliceTable[parm][time].num1 = num1;
   VSliceTable[parm][time].vx1 = cvx1;
   VSliceTable[parm][time].vy1 = cvy1;
   VSliceTable[parm][time].vz1 = cvz1;
   VSliceTable[parm][time].num2 = num2;
   VSliceTable[parm][time].vx2 = cvx2;
   VSliceTable[parm][time].vy2 = cvy2;
   VSliceTable[parm][time].vz2 = cvz2;
   VSliceTable[parm][time].num3 = num3;
   VSliceTable[parm][time].vx3 = cvx3;
   VSliceTable[parm][time].vy3 = cvy3;
   VSliceTable[parm][time].vz3 = cvz3;
   VSliceTable[parm][time].valid = 1;

   clear_lock( &VSliceTable[parm][time].lock );

   Redraw=1;
}





/*** calc_chslice *****************************************************
   Calculate an horizontal colored slice and store it.
   Input:  time - the time step.
           parm - the parameter.
	   level - position of slice in [0..Nl-1].
	   tasknum - which cpu to execute on.
   Output:  resulting data values are saved in CHSliceTable.
**********************************************************************/
void calc_chslice( time, parm, level, tasknum )
int time, parm;
float level;
int tasknum;
{
   float *grid;
   float *slice;
   int upper, lower, above, below;
   float a, b, g1, g2, val, scale;
   int i, j, bytes;
   uint_1 *buffer;
   int realparm;

   if (Debug)
     printf("calc_chslice%d( %d, %d, %f )\n", tasknum, time, parm, level );

   realparm = CloneTable[parm];

   grid = get_grid( time, realparm );  /* read the grid */
   if (!grid)
      die("couldn't get grid");

   /* extract the 2-D array from 3-D grid */
   slice = extract_hslice( grid, level, 0 );

   release_grid( time, realparm, grid );

   /* scale values to [0,1] with missing = 100.0 */
   scale = 1.0 / (MaxVal[realparm]-MinVal[realparm]);
   for (i=0;i<Nr*Nc;i++) {
      if (slice[i] >= 1.0e30)
	 slice[i] = 100.0;
      else
	 slice[i] = (slice[i]-MinVal[realparm]) * scale;
   }

   /* allocate 2-D array to extract and the buffer array */
   bytes = Nr * Nc * sizeof(uint_1);
   buffer = (uint_1 *) allocate( bytes );
   if (!buffer)
      die("Out of memory in calc_chslice");

   /************************** Compress ******************************/
   compress_float_to_char( slice, buffer, Nr*Nc );

   deallocate( slice, -1 );

   /**************************** Store ********************************/
   recent( CHSLICE, parm );

   /* Store the new slice */
   wait_lock( &CHSliceTable[parm][time].lock );

   /* deallocate existing slice, if any */
   free_chslice( time, parm );

   /* store new slice */
   CHSliceTable[parm][time].level = level;
   CHSliceTable[parm][time].data = buffer;
   CHSliceTable[parm][time].valid = 1;

   clear_lock( &CHSliceTable[parm][time].lock );
   Redraw = 1;
}






/*** calc_cvslice ******************************************************
   Calculate an vertical colored slice and store it.
   Input:  time - the time step.
           parm - the parameter.
	   r1, c1 - position of 1st corner in [0..Nr-1],[0..Nc-1]
	   r2, c2 - position of 1st corner in [0..Nr-1],[0..Nc-1]
	   tasknum - which cpu to execute on.
   Output:  resulting poly-triangle strip is saved in CVSliceTable.
**********************************************************************/
void calc_cvslice( time, parm, r1,c1, r2,c2, tasknum )
int time, parm;
float r1,c1, r2,c2;
int tasknum;
{
   float *grid, *slice;
   uint_1 *buffer;
   int cols, rows, i, bytes;
   float scale;
   int realparm;

   realparm = CloneTable[parm];

   grid = get_grid( time, realparm );  /* read the grid */
   if (!grid)
      die("couldn't get grid");

   rows = Nl;
   cols = (Nr>Nc) ? Nr : Nc;  /* kludge */

   slice = extract_vslice( grid, r1,c1,r2,c2, rows, cols, 0 );

   release_grid( time, realparm, grid );

   /* scale values to [0,1] with missing = 100.0 */
   scale = 1.0 / (MaxVal[realparm]-MinVal[realparm]);
   for (i=0;i<rows*cols;i++) {
      if (slice[i] >= 1.0e30) {
	 slice[i] = 100.0;
      }
      else
	 slice[i] = (slice[i]-MinVal[realparm]) * scale;
   }

   bytes = rows * cols * sizeof( uint_1 );
   buffer = (uint_1 *) allocate( bytes );
   if (!buffer)
      die("Out of memory in calc_cvslice");

   /************************ Compress ********************************/
   compress_float_to_char( slice, buffer, rows*cols );

   deallocate( slice, -1 );

   /************************* Store **********************************/
   recent( CVSLICE, parm );

   /* Store the new slice */
   wait_lock( &CVSliceTable[parm][time].lock );

   /* deallocate existing slice, if any */
   free_cvslice( time, parm );

   /* store new slice */
   CVSliceTable[parm][time].r1 = r1;
   CVSliceTable[parm][time].c1 = c1;
   CVSliceTable[parm][time].r2 = r2;
   CVSliceTable[parm][time].c2 = c2;
   CVSliceTable[parm][time].rows = rows;
   CVSliceTable[parm][time].cols = cols;
   CVSliceTable[parm][time].data = buffer;
   CVSliceTable[parm][time].valid = 1;

   clear_lock( &CVSliceTable[parm][time].lock );

   Redraw=1;
}


#define CROSS( c, a, b )  { c[0] =  a[1]*b[2]-a[2]*b[1]; \
			    c[1] = -a[0]*b[2]+a[2]*b[0]; \
			    c[2] =  a[0]*b[1]-a[1]*b[0]; \
			  }

#define MAGNITUDE( a )    sqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2] )

#define NORMALIZE( v ) { float mag = 1.0/sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]); \
			 v[0] *= mag; \
		         v[1] *= mag; \
			 v[2] *= mag; \
		       }

#define SCALE( v, d )  {  v[0] = v[0] / d; \
			  v[1] = v[1] / d; \
			  v[2] = v[2] / d; \
		       }

#define ARROW_SCALE  0.01


void calc_hwindslice( time, ws, level, scale, density, tasknum )
int time, ws;
float level, scale, density;
int tasknum;
{
   float *grid, *u, *v, *w, uu, vv, ww;
   float xscale, yscale, zscale;
   float xm, ym, zm;
   float midlat;
   int i, j, di, dj, n, deltatime, bytes;
   float vx[MAX_VERTS], vy[MAX_VERTS], vz[MAX_VERTS];
   float x,y,z, dx,dy;
   int max = MAX_VERTS;
   char buffer[2*MAX_VERTS];
   int_2 *cvx, *cvy, *cvz;

   if (!WindAvailable)
      return;

   if (scale<0.0)
      scale = 1.0;

   if (Debug)
      printf("calc_hwind( %d, %d, %f, %f )\n", time, ws, level, density );

   /* allocate 2-D arrays */
   grid = get_grid( time, Uparm );
   u = extract_hslice( grid, level, 0 );
   release_grid( time, Uparm, grid );
   grid = get_grid( time, Vparm );
   v = extract_hslice( grid, level, 0 );
   release_grid( time, Vparm, grid );
   grid = get_grid( time, Wparm );
   w = extract_hslice( grid, level, 0 );
   release_grid( time, Wparm, grid );

   /* size of box in meters */
   midlat = (NorthLat[time]+SouthLat[time]) / 2.0;
   xm = (WestLon[time]-EastLon[time]) * 111120.0 * cos( DEG2RAD(midlat) );
   ym = (NorthLat[time]-SouthLat[time]) * 111120.0;
   zm = (TopHgt[time]-BottomHgt[time]) * 1000.0;

   /* scaling from m/s to graphics/timestep */
   if (time>1)
      deltatime = Elapsed[time] - Elapsed[time-1];
   else if (NumTimes==1)
      deltatime = 600;  /* only 1 time step, let deltatime = 10 minutes */
   else
      deltatime = Elapsed[1] - Elapsed[0];

   xscale = 2.0 / xm * deltatime * scale;
   yscale = 2.0 / ym * deltatime * scale;
   zscale = 2.0 / zm * deltatime * scale;

   dx = 2.0 / (float) (Nc-1);
   dy = -2.0 / (float) (Nr-1);
   z = 1.0 - 2.0 * level / (float) (Nl-1);

   /* density calculations: */
   if (density>1.0 || density<0.0)
     density = 1.0;
   di = (int) (1.0 / density);
   dj = (int) (1.0 / density);
   dx = dx * (float) dj;
   dy = dy * (float) di;

   /* calculate vector vertices */
   n = 0;
   y = 1.0;
   for (i=0;i<Nr && n+4<MAX_VERTS;i+=di) {
      x = -1.0;
      for (j=0;j<Nc && n+4<MAX_VERTS;j+=dj,x+=dx) {
	 float dir[3];  /* wind vector direction */
	 float up[3], a[3];  /* up & across vectors */
	 float px, py, pz, len;

	 uu = u[i*Nc+j];
	 if (uu>1000.0 || uu<-1000.0)  continue;
	 vv = v[i*Nc+j];
	 if (vv>1000.0 || vv<-1000.0)  continue;
	 ww = w[i*Nc+j];
	 if (ww>1000.0 || ww<-1000.0)  continue;
	 dir[0] = uu * xscale;
	 dir[1] = vv * yscale;
	 dir[2] = -ww * zscale;
	 len = MAGNITUDE( dir );
	 if (len<=0.0)  continue;

	 /* main arrow vector */
	 vx[n] = x;
	 vy[n] = y;
	 vz[n] = z;
	 n++;
	 vx[n] = x + dir[0];
	 vy[n] = y + dir[1];
	 vz[n] = z + dir[2];
	 n++;

	 /* arrow head */
	 SCALE( dir, len );
	 if (dir[2]>0.5 || dir[2]<-0.5) {
	    up[0] = 0.0;  up[1] = 1.0;  up[2] = 0.0;
	 }
	 else {
	    up[0] = 0.0;  up[1] = 0.0;  up[2] = 1.0;
	 }
	 CROSS( a, dir, up );
	 a[0] *= ARROW_SCALE;
	 a[1] *= ARROW_SCALE;
	 a[2] *= ARROW_SCALE;
	 px = vx[n-1] - dir[0] * ARROW_SCALE;
	 py = vy[n-1] - dir[1] * ARROW_SCALE;
	 pz = vz[n-1] - dir[2] * ARROW_SCALE;
	 vx[n] = px + a[0];
	 vy[n] = py + a[1];
	 vz[n] = pz + a[2];
	 n++;
	 vx[n] = px - a[0];
	 vy[n] = py - a[1];
	 vz[n] = pz - a[2];
	 n++;

      }
      y += dy;
   }

   /* here, n = number of vertices */

   /* deallocate buffers */
   deallocate( u, -1 );
   deallocate( v, -1 );
   deallocate( w, -1 );

   /************************ Compress ********************************/

   if (n) {
      bytes = n*sizeof(int_2);
      cvx = (int_2 *) allocate( bytes );
      cvy = (int_2 *) allocate( bytes );
      cvz = (int_2 *) allocate( bytes );
      compress_vertices( vx, cvx, n );
      compress_vertices( vy, cvy, n );
      compress_vertices( vz, cvz, n );
   }
   else {
      cvx = cvy = cvz = NULL;
   }

   /************************* Store wind slice ***********************/
   /* Store the new wind slice */
   recent( HWIND, ws );

   wait_lock( &HWindTable[ws][time].lock );

   /* deallocate existing slice, if any */
   free_hwind( time, ws );

   /* store new slice */
   HWindTable[ws][time].level = level;
   HWindTable[ws][time].density = density;
   HWindTable[ws][time].scale = scale;
   HWindTable[ws][time].nvectors = n / 4;    /* 4 vertices / vector */
   HWindTable[ws][time].vx = cvx;
   HWindTable[ws][time].vy = cvy;
   HWindTable[ws][time].vz = cvz;
   HWindTable[ws][time].valid = 1;

   clear_lock( &HWindTable[ws][time].lock );

   Redraw = 1;
}



void calc_vwindslice( time, ws, r1,c1, r2,c2, scale, density, tasknum )
int time, ws;
float r1,c1, r2,c2, scale, density;
int tasknum;
{
   float *grid, *u, *v, *w, uu, vv, ww;
   int i, j, di, dj, n, rows, cols;
   float vx[MAX_VERTS], vy[MAX_VERTS], vz[MAX_VERTS];
   float x,y,z;
   float xscale, yscale, zscale, xm, ym, zm, midlat;
   int deltatime, bytes;
   float x0,y0,z0, x1,y1,z1, dx,dy,dz;
   int max = MAX_VERTS;
   char buffer[2*MAX_VERTS];
   int_2 *cvx, *cvy, *cvz;

   if (!WindAvailable)
      return;

   if (Debug)
      printf("calc_vwind( %d, %d, %f )\n", time, ws, density );

   if (scale<0.0)
      scale = 1.0;

   rows = Nl;
   cols = MAX(Nr,Nc);

   /* get u, v, w slices */
   grid = get_grid( time, Uparm );
   u = extract_vslice( grid, r1,c1, r2,c2, rows, cols, 0 );
   release_grid( time, Uparm, grid );
   grid = get_grid( time, Vparm );
   v = extract_vslice( grid, r1,c1, r2,c2, rows, cols, 0 );
   release_grid( time, Vparm, grid );
   grid = get_grid( time, Wparm );
   w = extract_vslice( grid, r1,c1, r2,c2, rows, cols, 0 );
   release_grid( time, Wparm, grid );

   /* size of box in meters */
   midlat = (NorthLat[time]+SouthLat[time]) / 2.0;
   xm = (WestLon[time]-EastLon[time]) * 111120.0 * cos( DEG2RAD(midlat) );
   ym = (NorthLat[time]-SouthLat[time]) * 111120.0;
   zm = (TopHgt[time]-BottomHgt[time]) * 1000.0;

   /* scaling from m/s to graphics/timestep */
   if (time>1)
      deltatime = Elapsed[time] - Elapsed[time-1];
   else if (NumTimes==1)
      deltatime = 600;  /* only 1 time step, let deltatime = 10 minutes */
   else
      deltatime = Elapsed[1] - Elapsed[0];
   if (NumTimes==1)
      deltatime = 1.0;
   xscale = 2.0 / xm * deltatime * scale;
   yscale = 2.0 / ym * deltatime * scale;
   zscale = 2.0 / zm * deltatime * scale;

   /* slice corners in graphics coords */
   x0 = -1.0 + 2.0 * c1 / (float) (Nc-1);
   y0 =  1.0 - 2.0 * r1 / (float) (Nr-1);
   z0 =  1.0;
   x1 = -1.0 + 2.0 * c2 / (float) (Nc-1);
   y1 =  1.0 - 2.0 * r2 / (float) (Nr-1);
   z1 = -1.0;

   dx = (x1-x0) / (float) (cols-1);
   dy = (y1-y0) / (float) (cols-1);
   dz = (z1-z0) / (float) (rows-1);

   /* density calculations: */
   if (density>1.0 || density<0.0)
     density = 1.0;
   di = (int) (1.0 / density);
   dj = (int) (1.0 / density);
   dx = dx * (float) dj;
   dy = dy * (float) dj;
   dz = dz * (float) di;

   n = 0;
   z = z0;
   for (i=0;i<rows && n+4<MAX_VERTS;i+=di) {
      x = x0;
      y = y0;
      for (j=0;j<cols && n+4<MAX_VERTS;j+=dj,x+=dx,y+=dy) {
	 float dir[3];  /* wind vector direction */
	 float up[3], a[3];  /* up & across vectors */
	 float px, py, pz, len;

	 uu = u[i*cols+j];
	 if (uu>1000.0 || uu<-1000.0)  continue;
	 vv = v[i*cols+j];
	 if (vv>1000.0 || vv<-1000.0)  continue;
	 ww = w[i*cols+j];
	 if (ww>1000.0 || ww<-1000.0)  continue;
	 dir[0] = uu * xscale;
	 dir[1] = vv * yscale;
	 dir[2] = -ww * zscale;
	 len = MAGNITUDE( dir );
	 if (len<=0.0)  continue;

	 /* main arrow vector */
	 vx[n] = x;
	 vy[n] = y;
	 vz[n] = z;
	 n++;

	 vx[n] = x + dir[0];
	 vy[n] = y + dir[1];
	 vz[n] = z + dir[2];
	 n++;

	 /* arrow head */
	 SCALE( dir, len );
	 up[0] = dy;  up[1] = dx;  up[2] = 0.0;
         NORMALIZE( up );
	 CROSS( a, dir, up );
	 a[0] *= ARROW_SCALE;
	 a[1] *= ARROW_SCALE;
	 a[2] *= ARROW_SCALE;
	 px = vx[n-1] - dir[0] * ARROW_SCALE;
	 py = vy[n-1] - dir[1] * ARROW_SCALE;
	 pz = vz[n-1] - dir[2] * ARROW_SCALE;
	 vx[n] = px + a[0];
	 vy[n] = py + a[1];
	 vz[n] = pz + a[2];
	 n++;
	 vx[n] = px - a[0];
	 vy[n] = py - a[1];
	 vz[n] = pz - a[2];
	 n++;

      }
      z += dz;
   }

   /* here, n = number of vertices */

   /* deallocate buffers */
   deallocate( u, -1 );
   deallocate( v, -1 );
   deallocate( w, -1 );

   /*************************** Compress *******************************/

   if (n) {
      bytes = n*sizeof(int_2);
      cvx = (int_2 *) allocate( bytes );
      cvy = (int_2 *) allocate( bytes );
      cvz = (int_2 *) allocate( bytes );
      compress_vertices( vx, cvx, n );
      compress_vertices( vy, cvy, n );
      compress_vertices( vz, cvz, n );
   }
   else {
      cvx = cvy = cvz = NULL;
   }

   /************************ Store wind slice ***************************/

   /* Store the new wind slice */
   recent( VWIND, ws );

   wait_lock( &VWindTable[ws][time].lock );

   /* deallocate existing slice, if any */
   free_vwind( time, ws );

   /* store new slice */
   VWindTable[ws][time].r1 = r1;
   VWindTable[ws][time].c1 = c1;
   VWindTable[ws][time].r2 = r2;
   VWindTable[ws][time].c2 = c2;
   VWindTable[ws][time].density = density;
   VWindTable[ws][time].scale = scale;
   VWindTable[ws][time].nvectors = n/4;   /* 4 vertices / vector */
   VWindTable[ws][time].vx = cvx;
   VWindTable[ws][time].vy = cvy;
   VWindTable[ws][time].vz = cvz;
   VWindTable[ws][time].valid = 1;

   clear_lock( &VWindTable[ws][time].lock );

   Redraw = 1;
}


/* maximum trajectory vertices */
#define MAX_TRAJV 5000


/*** calc_traj ********************************************************
   Calculate a trajectory.
   Input:  row, col, lev - start position in grid coords.
           time - start time in [0..NumTimes-1].
           id - trajectory identification.
   Output:  list of coordinates and times are saved in TrajTable.
**********************************************************************/
void calc_traj( row, col, lev, time, id, ribbon, step, tlen )
float row, col, lev;
int time, id, ribbon;
float step, tlen;
{
   int len, i;
   float vx[MAX_TRAJV], vy[MAX_TRAJV], vz[MAX_TRAJV];  /* vertices */
   float nx[MAX_TRAJV], ny[MAX_TRAJV], nz[MAX_TRAJV];  /* normals */
   int tt[MAX_TRAJV];  /* traj times */
   char buffer[MAX_TRAJV*4];
   float xs, ys, zs;
   int ifrlen, bytes, t;
   int_2 *cvx, *cvy, *cvz;
   int_1 *cnx, *cny, *cnz;
   int ictl, icth, icta, ictb;
   int itb, ita;

   if (NumTraj>=MAXTRAJ) {
      /* out of trajectory space */
      return;
   }

   if (Debug)
      printf("calc_traj( %f %f %f %d %d )\n", row, col, lev, time, id );
   len = trace( row, col, lev, time, step, tlen,
	       MAX_TRAJV, vy,vx,vz, tt, &ifrlen );

   if (len==0)
      return;

   /* convert coords from grid to graphics in [-1,1]^3 */
   xs = 2.0 / (float) (Nc-1);
   ys = 2.0 / (float) (Nr-1);
   zs = 2.0 / (float) (Nl-1);
   for (i=0;i<len;i++) {  /* stellar vectorized */
      vx[i] = vx[i] * xs - 1.0;
      vy[i] = 1.0 - vy[i] * ys;
      vz[i] = 1.0 - vz[i] * zs;
/*  printf("%d: vx,vy,vz,tt = %f %f %f %d\n", i, vx[i], vy[i], vz[i], tt[i]);*/
   }

   if (ribbon) {
      /* convert to ribbon, calc normals */
      len = to_ribbon( len, vx,vy,vz, tt,  nx,ny,nz );
   }
   if (len==0)
      return;

   if (Debug) {
      if (ribbon)
	printf("sending ribbon, len=%d ifrlen=%d...", len, ifrlen);
      else
	printf("sending traj, len=%d ifrlen=%d...", len, ifrlen);
   }

   /****************************** Compress ***************************/
   bytes = len * sizeof(int_2);
   cvx = (int_2 *) allocate( bytes );
   cvy = (int_2 *) allocate( bytes );
   cvz = (int_2 *) allocate( bytes );
   assert( cvx && cvy && cvz );
   compress_vertices( vx, cvx, len );
   compress_vertices( vy, cvy, len );
   compress_vertices( vz, cvz, len );

   if (ribbon) {
      bytes = len * sizeof(int_1);
      cnx = (int_1 *) allocate( bytes );
      cny = (int_1 *) allocate( bytes );
      cnz = (int_1 *) allocate( bytes );
      compress_normals( nx, cnx, len );
      compress_normals( ny, cny, len );
      compress_normals( nz, cnz, len );
   }
   else {
      cnx = cny = cnz = NULL;
   }

   /***************************** Store ******************************/

   recent( TRAJ, id );

   LOCK_ON( Gfx_Lock );
   t = NumTraj;
   NumTraj++;
   LOCK_OFF( Gfx_Lock );

   wait_lock( &TrajTable[t].lock );

   TrajTable[t].length = len;
   TrajTable[t].vx = cvx;
   TrajTable[t].vy = cvy;
   TrajTable[t].vz = cvz;
   TrajTable[t].nx = cnx;
   TrajTable[t].ny = cny;
   TrajTable[t].nz = cnz;
   TrajTable[t].start = (uint_2 *) allocate( NumTimes * sizeof(uint_2) );
   TrajTable[t].len = (uint_2 *) allocate( NumTimes * sizeof(uint_2) );

   /* calculate start and len array values */
   if (ifrlen==0)
     ifrlen = 3;   /* FIX THIS */

   ictl = 1;
   icth = len;
   icta = ictb = ictl;

   for (itb=1;itb<=NumTimes;itb++) {
      ita = itb-ifrlen;
      while ( !(tt[icta-1]>=ita || icta>=icth) )
	icta++;
      while ( !(tt[ictb-1]>=itb || ictb>=icth) )
	ictb++;
      if (ictb==ictl) {
	 TrajTable[t].start[itb-1] = 0xffff;
      }
      else {
	 TrajTable[t].start[itb-1] = icta;  /*-1;*/
	 TrajTable[t].len[itb-1] = ictb-icta;
      }
/*
      printf("start[%d] = %d  len = %d\n", itb-1,
	     TrajTable[t].start[itb-1],
	     TrajTable[t].len[itb-1] );
*/
   }

   TrajTable[t].group = id;
   TrajTable[t].kind = ribbon;

   clear_lock( &TrajTable[t].lock );

   Redraw = 1;
}





/*** compress_vertices ************************************************
   Convert an array of floats in [-1,1] to 2-byte scaled integers
   in [-10000,10000].  This is used to compress arrays of vertices.
   Input: from - pointer to input array of floats
          to - pointer to output array of short ints
          num - number of elements in array.
   Output:  to - result of conversion.
**********************************************************************/
static void compress_vertices( from, to, num )
float from[];
int_2 to[];
int num;
{
   int i;
   /*$dir no_recurrence  */  /* fully vectorized */
   for (i=0;i<num;i++) {
      to[i] = (int_2) (from[i]*10000.0);
   }
}




/*** compress_normals *************************************************
   Compress an array of floating point numbers in [-1,1] to 1-byte
   scaled integers in [-125,125].  This is typically used to compress
   normal vectors.
   Input:  from - pointer to input array of floats
           to - pointer to output array of chars
           num - number of elements
   Output:  to - result of conversion
**********************************************************************/
static void compress_normals( from, to, num )
float from[];
int_1 to[];
int num;
{
   int i;

   /*$dir no_recurrence */
   for (i=0; i<num; i++) {
      to[i] = (int_1) (from[i]*125.0);
   }
}



/*** compress_float_to_char *******************************************
   Convert an array of floats in the interval [0,1] to
   an array of bytes in the interval [0..254].  Values outside [0,1]
   will be stored as 255.
   Input: from - pointer to input array of floats
          to - pointer to output array of chars/bytes.
          cnt - number of elements in array.
   Output:  to - result of conversion.
**********************************************************************/
static void compress_float_to_char( from, to, cnt )
float from[];
uint_1 *to;
int cnt;
{
   int i;

   for (i=0; i<cnt; i++) {
      if (from[i]>1.0) {
	 *to++ = (uint_1) 0xff;
      }
      else {
	 *to++ = (uint_1) (from[i]*254.0);
      }
   }
}




/*** exdig ***********************************************************
   Extract digits from a float to be used by the PLTDG subroutine
   in zcont.f
   Input:  x - the floating point number
   Output: dig - number of digits to plot
           j   - integer to left of decimal point
           j1, j2, j3 - integer digits to right of decimal point
           sign - the sign of the number: 1 or -1
**********************************************************************/
#ifdef ibm
  exdig( x, dig, j, j1, j2, j3, sign )
#else
  exdig_( x, dig, j, j1, j2, j3, sign )
#endif
float *x, *dig;
int *j, *j1, *j2, *j3;
int *sign;
{
   float y;

   *j = (int) *x;
   if (*j<0) {
      *j = -*j;
      y = -*x;
      *sign = -1;
      *dig = 0.5;
   }
   else {
      y = *x;
      *sign = 1;
      *dig = 0.0;
   }

   *j1 = ( (int) (y * 10.0) ) % 10;
   *j2 = ( (int) (y * 100.0) ) % 10;
   *j3 = ( (int) (y * 1000.0) ) % 10;

   /* examine digits to left of decimal point */
   if (*j>=100) {
      *j1 = *j2 = *j3 = 0;
      *dig += 3.0;
   }
   else if (*j>=10) {
      *j3 = 0;
      *dig += 4.5;
      if (*j2==0) {
         *dig -= 1.0;
         if (*j1==0)  *dig -= 1.0;
      }
   }
   else {
      *dig += 4.5;
      if (*j3==0) {
         *dig -= 1.0;
         if (*j2==0) {
            *dig -= 1.0;
            if (*j1==0)  *dig -= 1.0;
         }
      }
   }

}
